use std::collections::BTreeMap;

use itertools::Itertools;

use crate::data::{Root, SourceFile, Span, Spanned};

use super::{LedgerEntry, Transaction};

fn print_transaction(root: &Root, tx: &Transaction, padding: usize) {
    println!(
        "{}-{:02}-{:02}: {}",
        tx.datestamp.0,
        tx.datestamp.1,
        tx.datestamp.2,
        tx.title.as_deref().unwrap_or("")
    );

    if !tx.annotations.is_empty() {
        println!(
            "    {}",
            tx.annotations.iter().map(|a| format!("[{a}]")).join(" ")
        );
    }

    let force_show = tx.balances.iter().unique_by(|b| *b.account).count() > 1;

    for bal in &tx.balances {
        let spacer = if bal.amount.is_sign_positive() {
            " "
        } else {
            ""
        };
        if Some(bal.unit.as_ref()) == root.account_spec(*bal.account).unwrap().unit.as_ref()
            && !force_show
        {
            println!(" - {:padding$}: {spacer}{}", bal.account, bal.amount,);
        } else {
            println!(
                " - {:padding$}: {spacer}{} {}",
                bal.account, bal.amount, bal.unit,
            );
        }
    }
    // empty line afterwards
    println!();
}

fn print_comment(c: &Spanned<String>) {
    println!("{c}");
}

pub fn print_ledger<'l>(root: &Root, entries: impl Iterator<Item = &'l LedgerEntry>) {
    let mut ordering = BTreeMap::<SourceFile, BTreeMap<Span, &LedgerEntry>>::new();

    entries.for_each(|e| {
        ordering
            .entry(e.span().context)
            .or_default()
            .insert(e.span(), &e);
    });

    let Some(padding) = root.spec_root.accounts.keys().map(|k| k.len()).max() else {
        // no accounts
        return;
    };

    for (filename, entries) in ordering {
        println!(
            "==== file {} ====",
            std::path::Path::new(filename.as_ref()).display()
        );
        for (_span, le) in entries {
            match le {
                LedgerEntry::Transaction(tx) => print_transaction(root, tx, padding),
                LedgerEntry::Comment(c) => print_comment(c),
            }
        }
    }
}
